@@ -30,14 +30,14 @@
 #if ENABLE(WEBASSEMBLY)
 
 #include "Options.h"
+#include <wtf/CheckedArithmetic.h>
 #include <wtf/DataLog.h>
 #include <wtf/Gigacage.h>
 #include <wtf/Lock.h>
-#include <wtf/OSAllocator.h>
-#include <wtf/PageBlock.h>
 #include <wtf/Platform.h>
 #include <wtf/PrintStream.h>
 #include <wtf/RAMSize.h>
+#include <wtf/StdSet.h>
 #include <wtf/Vector.h>
 
 #include <cstring>
@@ -105,7 +105,7 @@ public:
     MemoryResult tryAllocateFastMemory()
     {
         MemoryResult result = [&] {
-            auto holder = holdLock(m_lock);
+            Locker locker { m_lock };
             if (m_fastMemories.size() >= m_maxFastMemoryCount)
                 return MemoryResult(nullptr, MemoryResult::SyncTryToReclaimMemory);
             
@@ -128,23 +128,62 @@ public:
     void freeFastMemory(void* basePtr)
     {
         {
-            auto holder = holdLock(m_lock);
+            Locker locker { m_lock };
             Gigacage::freeVirtualPages(Gigacage::Primitive, basePtr, Memory::fastMappedBytes());
             m_fastMemories.removeFirst(basePtr);
         }
         
         dataLogLnIf(Options::logWebAssemblyMemory(), "Freed virtual; state: ", *this);
     }
-    
-    bool isAddressInFastMemory(void* address)
+
+    MemoryResult tryAllocateGrowableBoundsCheckingMemory(size_t mappedCapacity)
     {
-        // NOTE: This can be called from a signal handler, but only after we proved that we're in JIT code.
-        auto holder = holdLock(m_lock);
+        MemoryResult result = [&] {
+            Locker locker { m_lock };
+            void* result = Gigacage::tryAllocateZeroedVirtualPages(Gigacage::Primitive, mappedCapacity);
+            if (!result)
+                return MemoryResult(nullptr, MemoryResult::SyncTryToReclaimMemory);
+
+            m_growableBoundsCheckingMemories.insert(std::make_pair(bitwise_cast<uintptr_t>(result), mappedCapacity));
+
+            return MemoryResult(result, MemoryResult::Success);
+        }();
+
+        dataLogLnIf(Options::logWebAssemblyMemory(), "Allocated virtual: ", result, "; state: ", *this);
+
+        return result;
+    }
+
+    void freeGrowableBoundsCheckingMemory(void* basePtr, size_t mappedCapacity)
+    {
+        {
+            Locker locker { m_lock };
+            Gigacage::freeVirtualPages(Gigacage::Primitive, basePtr, mappedCapacity);
+            m_growableBoundsCheckingMemories.erase(std::make_pair(bitwise_cast<uintptr_t>(basePtr), mappedCapacity));
+        }
+
+        dataLogLnIf(Options::logWebAssemblyMemory(), "Freed virtual; state: ", *this);
+    }
+
+    bool isInGrowableOrFastMemory(void* address)
+    {
+        // NOTE: This can be called from a signal handler, but only after we proved that we're in JIT code or WasmLLInt code.
+        Locker locker { m_lock };
         for (void* memory : m_fastMemories) {
             char* start = static_cast<char*>(memory);
             if (start <= address && address <= start + Memory::fastMappedBytes())
                 return true;
         }
+        uintptr_t addressValue = bitwise_cast<uintptr_t>(address);
+        auto iterator = std::upper_bound(m_growableBoundsCheckingMemories.begin(), m_growableBoundsCheckingMemories.end(), std::make_pair(addressValue, 0),
+            [](std::pair<uintptr_t, size_t> a, std::pair<uintptr_t, size_t> b) {
+                return (a.first + a.second) < (b.first + b.second);
+            });
+        if (iterator != m_growableBoundsCheckingMemories.end()) {
+            // Since we never have overlapped range in m_growableBoundsCheckingMemories, just checking one lower-bound range is enough.
+            if (iterator->first <= addressValue && addressValue < (iterator->first + iterator->second))
+                return true;
+        }
         return false;
     }
 
@@ -158,7 +197,7 @@ public:
     MemoryResult::Kind tryAllocatePhysicalBytes(size_t bytes)
     {
         MemoryResult::Kind result = [&] {
-            auto holder = holdLock(m_lock);
+            Locker locker { m_lock };
             if (m_physicalBytes + bytes > memoryLimit())
                 return MemoryResult::SyncTryToReclaimMemory;
             
@@ -178,7 +217,7 @@ public:
     void freePhysicalBytes(size_t bytes)
     {
         {
-            auto holder = holdLock(m_lock);
+            Locker locker { m_lock };
             m_physicalBytes -= bytes;
         }
         
@@ -194,6 +233,7 @@ private:
     Lock m_lock;
     unsigned m_maxFastMemoryCount { 0 };
     Vector<void*> m_fastMemories;
+    StdSet<std::pair<uintptr_t, size_t>> m_growableBoundsCheckingMemories;
     size_t m_physicalBytes { 0 };
 };
 
@@ -237,30 +277,82 @@ bool tryAllocate(const Func& allocate, const WTF::Function<void(Memory::NotifyPr
 
 } // anonymous namespace
 
+
+MemoryHandle::MemoryHandle(void* memory, size_t size, size_t mappedCapacity, PageCount initial, PageCount maximum, MemorySharingMode sharingMode, MemoryMode mode)
+    : m_sharingMode(sharingMode)
+    , m_mode(mode)
+    , m_memory(memory, mappedCapacity)
+    , m_size(size)
+    , m_mappedCapacity(mappedCapacity)
+    , m_initial(initial)
+    , m_maximum(maximum)
+{
+#if ASSERT_ENABLED
+    if (sharingMode == MemorySharingMode::Default && mode == MemoryMode::BoundsChecking)
+        ASSERT(mappedCapacity == size);
+#endif
+}
+
+MemoryHandle::~MemoryHandle()
+{
+    if (m_memory) {
+        void* memory = this->memory();
+        memoryManager().freePhysicalBytes(m_size);
+        switch (m_mode) {
+        case MemoryMode::Signaling:
+            if (mprotect(memory, Memory::fastMappedBytes(), PROT_READ | PROT_WRITE)) {
+                dataLog("mprotect failed: ", strerror(errno), "\n");
+                RELEASE_ASSERT_NOT_REACHED();
+            }
+            memoryManager().freeFastMemory(memory);
+            break;
+        case MemoryMode::BoundsChecking: {
+            switch (m_sharingMode) {
+            case MemorySharingMode::Default:
+                Gigacage::freeVirtualPages(Gigacage::Primitive, memory, m_size);
+                break;
+            case MemorySharingMode::Shared: {
+                if (mprotect(memory, m_mappedCapacity, PROT_READ | PROT_WRITE)) {
+                    dataLog("mprotect failed: ", strerror(errno), "\n");
+                    RELEASE_ASSERT_NOT_REACHED();
+                }
+                memoryManager().freeGrowableBoundsCheckingMemory(memory, m_mappedCapacity);
+                break;
+            }
+            }
+            break;
+        }
+        }
+    }
+}
+
+// FIXME: ARM64E clang has a bug and inlining this function makes optimizer run forever.
+// For now, putting NEVER_INLINE to suppress inlining of this.
+NEVER_INLINE void* MemoryHandle::memory() const
+{
+    ASSERT(m_memory.getMayBeNull(m_mappedCapacity) == m_memory.getUnsafe());
+    return m_memory.getMayBeNull(m_mappedCapacity);
+}
+
 Memory::Memory()
+    : m_handle(adoptRef(*new MemoryHandle(nullptr, 0, 0, PageCount(0), PageCount(0), MemorySharingMode::Default, MemoryMode::BoundsChecking)))
 {
 }
 
-Memory::Memory(PageCount initial, PageCount maximum, Function<void(NotifyPressure)>&& notifyMemoryPressure, Function<void(SyncTryToReclaim)>&& syncTryToReclaimMemory, WTF::Function<void(GrowSuccess, PageCount, PageCount)>&& growSuccessCallback)
-    : m_initial(initial)
-    , m_maximum(maximum)
+Memory::Memory(PageCount initial, PageCount maximum, MemorySharingMode sharingMode, Function<void(NotifyPressure)>&& notifyMemoryPressure, Function<void(SyncTryToReclaim)>&& syncTryToReclaimMemory, WTF::Function<void(GrowSuccess, PageCount, PageCount)>&& growSuccessCallback)
+    : m_handle(adoptRef(*new MemoryHandle(nullptr, 0, 0, initial, maximum, sharingMode, MemoryMode::BoundsChecking)))
     , m_notifyMemoryPressure(WTFMove(notifyMemoryPressure))
     , m_syncTryToReclaimMemory(WTFMove(syncTryToReclaimMemory))
     , m_growSuccessCallback(WTFMove(growSuccessCallback))
 {
     ASSERT(!initial.bytes());
-    ASSERT(m_mode == MemoryMode::BoundsChecking);
+    ASSERT(mode() == MemoryMode::BoundsChecking);
     dataLogLnIf(verbose, "Memory::Memory allocating ", *this);
     ASSERT(!memory());
 }
 
-Memory::Memory(void* memory, PageCount initial, PageCount maximum, size_t mappedCapacity, MemoryMode mode, Function<void(NotifyPressure)>&& notifyMemoryPressure, Function<void(SyncTryToReclaim)>&& syncTryToReclaimMemory, WTF::Function<void(GrowSuccess, PageCount, PageCount)>&& growSuccessCallback)
-    : m_memory(memory, initial.bytes())
-    , m_size(initial.bytes())
-    , m_initial(initial)
-    , m_maximum(maximum)
-    , m_mappedCapacity(mappedCapacity)
-    , m_mode(mode)
+Memory::Memory(Ref<MemoryHandle>&& handle, Function<void(NotifyPressure)>&& notifyMemoryPressure, Function<void(SyncTryToReclaim)>&& syncTryToReclaimMemory, WTF::Function<void(GrowSuccess, PageCount, PageCount)>&& growSuccessCallback)
+    : m_handle(WTFMove(handle))
     , m_notifyMemoryPressure(WTFMove(notifyMemoryPressure))
     , m_syncTryToReclaimMemory(WTFMove(syncTryToReclaimMemory))
     , m_growSuccessCallback(WTFMove(growSuccessCallback))
@@ -273,7 +365,12 @@ Ref<Memory> Memory::create()
     return adoptRef(*new Memory());
 }
 
-RefPtr<Memory> Memory::tryCreate(PageCount initial, PageCount maximum, WTF::Function<void(NotifyPressure)>&& notifyMemoryPressure, WTF::Function<void(SyncTryToReclaim)>&& syncTryToReclaimMemory, WTF::Function<void(GrowSuccess, PageCount, PageCount)>&& growSuccessCallback)
+Ref<Memory> Memory::create(Ref<MemoryHandle>&& handle, WTF::Function<void(NotifyPressure)>&& notifyMemoryPressure, WTF::Function<void(SyncTryToReclaim)>&& syncTryToReclaimMemory, WTF::Function<void(GrowSuccess, PageCount, PageCount)>&& growSuccessCallback)
+{
+    return adoptRef(*new Memory(WTFMove(handle), WTFMove(notifyMemoryPressure), WTFMove(syncTryToReclaimMemory), WTFMove(growSuccessCallback)));
+}
+
+RefPtr<Memory> Memory::tryCreate(PageCount initial, PageCount maximum, MemorySharingMode sharingMode, WTF::Function<void(NotifyPressure)>&& notifyMemoryPressure, WTF::Function<void(SyncTryToReclaim)>&& syncTryToReclaimMemory, WTF::Function<void(GrowSuccess, PageCount, PageCount)>&& growSuccessCallback)
 {
     ASSERT(initial);
     RELEASE_ASSERT(!maximum || maximum >= initial); // This should be guaranteed by our caller.
@@ -287,7 +384,7 @@ RefPtr<Memory> Memory::tryCreate(PageCount initial, PageCount maximum, WTF::Func
     if (maximum && !maximumBytes) {
         // User specified a zero maximum, initial size must also be zero.
         RELEASE_ASSERT(!initialBytes);
-        return adoptRef(new Memory(initial, maximum, WTFMove(notifyMemoryPressure), WTFMove(syncTryToReclaimMemory), WTFMove(growSuccessCallback)));
+        return adoptRef(new Memory(initial, maximum, sharingMode, WTFMove(notifyMemoryPressure), WTFMove(syncTryToReclaimMemory), WTFMove(growSuccessCallback)));
     }
     
     bool done = tryAllocate(
@@ -308,47 +405,55 @@ RefPtr<Memory> Memory::tryCreate(PageCount initial, PageCount maximum, WTF::Func
     }
     
     if (fastMemory) {
-        
         if (mprotect(fastMemory + initialBytes, Memory::fastMappedBytes() - initialBytes, PROT_NONE)) {
             dataLog("mprotect failed: ", strerror(errno), "\n");
             RELEASE_ASSERT_NOT_REACHED();
         }
 
-        return adoptRef(new Memory(fastMemory, initial, maximum, Memory::fastMappedBytes(), MemoryMode::Signaling, WTFMove(notifyMemoryPressure), WTFMove(syncTryToReclaimMemory), WTFMove(growSuccessCallback)));
+        return Memory::create(adoptRef(*new MemoryHandle(fastMemory, initialBytes, Memory::fastMappedBytes(), initial, maximum, sharingMode, MemoryMode::Signaling)), WTFMove(notifyMemoryPressure), WTFMove(syncTryToReclaimMemory), WTFMove(growSuccessCallback));
     }
     
     if (UNLIKELY(Options::crashIfWebAssemblyCantFastMemory()))
         webAssemblyCouldntGetFastMemory();
 
-    if (!initialBytes)
-        return adoptRef(new Memory(initial, maximum, WTFMove(notifyMemoryPressure), WTFMove(syncTryToReclaimMemory), WTFMove(growSuccessCallback)));
-    
-    void* slowMemory = Gigacage::tryAllocateZeroedVirtualPages(Gigacage::Primitive, initialBytes);
-    if (!slowMemory) {
-        memoryManager().freePhysicalBytes(initialBytes);
-        return nullptr;
+    switch (sharingMode) {
+    case MemorySharingMode::Default: {
+        if (!initialBytes)
+            return adoptRef(new Memory(initial, maximum, sharingMode, WTFMove(notifyMemoryPressure), WTFMove(syncTryToReclaimMemory), WTFMove(growSuccessCallback)));
+
+        void* slowMemory = Gigacage::tryAllocateZeroedVirtualPages(Gigacage::Primitive, initialBytes);
+        if (!slowMemory) {
+            memoryManager().freePhysicalBytes(initialBytes);
+            return nullptr;
+        }
+        return Memory::create(adoptRef(*new MemoryHandle(slowMemory, initialBytes, initialBytes, initial, maximum, sharingMode, MemoryMode::BoundsChecking)), WTFMove(notifyMemoryPressure), WTFMove(syncTryToReclaimMemory), WTFMove(growSuccessCallback));
     }
-    return adoptRef(new Memory(slowMemory, initial, maximum, initialBytes, MemoryMode::BoundsChecking, WTFMove(notifyMemoryPressure), WTFMove(syncTryToReclaimMemory), WTFMove(growSuccessCallback)));
+    case MemorySharingMode::Shared: {
+        char* slowMemory = nullptr;
+        tryAllocate(
+            [&] () -> MemoryResult::Kind {
+                auto result = memoryManager().tryAllocateGrowableBoundsCheckingMemory(maximumBytes);
+                slowMemory = bitwise_cast<char*>(result.basePtr);
+                return result.kind;
+            }, notifyMemoryPressure, syncTryToReclaimMemory);
+        if (!slowMemory) {
+            memoryManager().freePhysicalBytes(initialBytes);
+            return nullptr;
+        }
+
+        if (mprotect(slowMemory + initialBytes, maximumBytes - initialBytes, PROT_NONE)) {
+            dataLog("mprotect failed: ", strerror(errno), "\n");
+            RELEASE_ASSERT_NOT_REACHED();
+        }
+
+        return Memory::create(adoptRef(*new MemoryHandle(slowMemory, initialBytes, maximumBytes, initial, maximum, sharingMode, MemoryMode::BoundsChecking)), WTFMove(notifyMemoryPressure), WTFMove(syncTryToReclaimMemory), WTFMove(growSuccessCallback));
+    }
+    }
+    RELEASE_ASSERT_NOT_REACHED();
+    return nullptr;
 }
 
-Memory::~Memory()
-{
-    if (m_memory) {
-        memoryManager().freePhysicalBytes(m_size);
-        switch (m_mode) {
-        case MemoryMode::Signaling:
-            if (mprotect(memory(), Memory::fastMappedBytes(), PROT_READ | PROT_WRITE)) {
-                dataLog("mprotect failed: ", strerror(errno), "\n");
-                RELEASE_ASSERT_NOT_REACHED();
-            }
-            memoryManager().freeFastMemory(memory());
-            break;
-        case MemoryMode::BoundsChecking:
-            Gigacage::freeVirtualPages(Gigacage::Primitive, memory(), m_size);
-            break;
-        }
-    }
-}
+Memory::~Memory() = default;
 
 size_t Memory::fastMappedRedzoneBytes()
 {
@@ -358,21 +463,80 @@ size_t Memory::fastMappedRedzoneBytes()
 size_t Memory::fastMappedBytes()
 {
     static_assert(sizeof(uint64_t) == sizeof(size_t), "We rely on allowing the maximum size of Memory we map to be 2^32 + redzone which is larger than fits in a 32-bit integer that we'd pass to mprotect if this didn't hold.");
-    return static_cast<size_t>(std::numeric_limits<uint32_t>::max()) + fastMappedRedzoneBytes();
+    return (static_cast<size_t>(1) << 32) + fastMappedRedzoneBytes();
 }
 
-bool Memory::addressIsInActiveFastMemory(void* address)
+bool Memory::addressIsInGrowableOrFastMemory(void* address)
 {
-    return memoryManager().isAddressInFastMemory(address);
+    return memoryManager().isInGrowableOrFastMemory(address);
+}
+
+Expected<PageCount, Memory::GrowFailReason> Memory::growShared(PageCount delta)
+{
+    Wasm::PageCount oldPageCount;
+    Wasm::PageCount newPageCount;
+    auto result = ([&]() -> Expected<PageCount, Memory::GrowFailReason> {
+        Locker locker { m_handle->lock() };
+
+        oldPageCount = sizeInPages();
+        newPageCount = oldPageCount + delta;
+        if (!newPageCount || !newPageCount.isValid())
+            return makeUnexpected(GrowFailReason::InvalidGrowSize);
+        if (newPageCount.bytes() > MAX_ARRAY_BUFFER_SIZE)
+            return makeUnexpected(GrowFailReason::OutOfMemory);
+
+        if (!delta.pageCount())
+            return oldPageCount;
+
+        dataLogLnIf(verbose, "Memory::grow(", delta, ") to ", newPageCount, " from ", *this);
+        RELEASE_ASSERT(newPageCount > PageCount::fromBytes(size()));
+
+        if (maximum() && newPageCount > maximum())
+            return makeUnexpected(GrowFailReason::WouldExceedMaximum);
+
+        size_t desiredSize = newPageCount.bytes();
+        RELEASE_ASSERT(desiredSize <= MAX_ARRAY_BUFFER_SIZE);
+        RELEASE_ASSERT(desiredSize > size());
+
+        // If the memory is MemorySharingMode::Shared, we already allocated enough virtual address space even if the memory is bound-checking mode. We perform mprotect to extend.
+        size_t extraBytes = desiredSize - size();
+        RELEASE_ASSERT(extraBytes);
+        bool allocationSuccess = tryAllocate(
+            [&] () -> MemoryResult::Kind {
+                return memoryManager().tryAllocatePhysicalBytes(extraBytes);
+            }, [](Wasm::Memory::NotifyPressure) { }, [](Memory::SyncTryToReclaim) { });
+        if (!allocationSuccess)
+            return makeUnexpected(GrowFailReason::OutOfMemory);
+
+        void* memory = this->memory();
+        RELEASE_ASSERT(memory);
+
+        // Signaling memory must have been pre-allocated virtually.
+        uint8_t* startAddress = static_cast<uint8_t*>(memory) + size();
+
+        dataLogLnIf(verbose, "Marking WebAssembly memory's ", RawPointer(memory), " as read+write in range [", RawPointer(startAddress), ", ", RawPointer(startAddress + extraBytes), ")");
+        if (mprotect(startAddress, extraBytes, PROT_READ | PROT_WRITE)) {
+            dataLog("mprotect failed: ", strerror(errno), "\n");
+            RELEASE_ASSERT_NOT_REACHED();
+        }
+
+        m_handle->growToSize(desiredSize);
+        return oldPageCount;
+    }());
+    if (result)
+        m_growSuccessCallback(GrowSuccessTag, oldPageCount, newPageCount);
+    return result;
 }
 
 Expected<PageCount, Memory::GrowFailReason> Memory::grow(PageCount delta)
 {
-    const Wasm::PageCount oldPageCount = sizeInPages();
-
     if (!delta.isValid())
         return makeUnexpected(GrowFailReason::InvalidDelta);
-    
+
+    if (sharingMode() == MemorySharingMode::Shared)
+        return growShared(delta);
+
+    const Wasm::PageCount oldPageCount = sizeInPages();
     const Wasm::PageCount newPageCount = oldPageCount + delta;
     if (!newPageCount || !newPageCount.isValid())
         return makeUnexpected(GrowFailReason::InvalidGrowSize);
@@ -393,52 +557,59 @@ Expected<PageCount, Memory::GrowFailReason> Memory::grow(PageCount delta)
         return success();
 
     dataLogLnIf(verbose, "Memory::grow(", delta, ") to ", newPageCount, " from ", *this);
-    RELEASE_ASSERT(newPageCount > PageCount::fromBytes(m_size));
+    RELEASE_ASSERT(newPageCount > PageCount::fromBytes(size()));
 
     if (maximum() && newPageCount > maximum())
         return makeUnexpected(GrowFailReason::WouldExceedMaximum);
 
     size_t desiredSize = newPageCount.bytes();
     RELEASE_ASSERT(desiredSize <= MAX_ARRAY_BUFFER_SIZE);
-    RELEASE_ASSERT(desiredSize > m_size);
-    size_t extraBytes = desiredSize - m_size;
-    RELEASE_ASSERT(extraBytes);
-    bool allocationSuccess = tryAllocate(
-        [&] () -> MemoryResult::Kind {
-            return memoryManager().tryAllocatePhysicalBytes(extraBytes);
-        }, m_notifyMemoryPressure, m_syncTryToReclaimMemory);
-    if (!allocationSuccess)
-        return makeUnexpected(GrowFailReason::OutOfMemory);
-
+    RELEASE_ASSERT(desiredSize > size());
     switch (mode()) {
     case MemoryMode::BoundsChecking: {
+        bool allocationSuccess = tryAllocate(
+            [&] () -> MemoryResult::Kind {
+                return memoryManager().tryAllocatePhysicalBytes(desiredSize);
+            }, m_notifyMemoryPressure, m_syncTryToReclaimMemory);
+        if (!allocationSuccess)
+            return makeUnexpected(GrowFailReason::OutOfMemory);
+
         RELEASE_ASSERT(maximum().bytes() != 0);
 
         void* newMemory = Gigacage::tryAllocateZeroedVirtualPages(Gigacage::Primitive, desiredSize);
         if (!newMemory)
             return makeUnexpected(GrowFailReason::OutOfMemory);
 
-        memcpy(newMemory, memory(), m_size);
-        if (m_memory)
-            Gigacage::freeVirtualPages(Gigacage::Primitive, memory(), m_size);
-        m_memory = CagedMemory(newMemory, desiredSize);
-        m_mappedCapacity = desiredSize;
-        m_size = desiredSize;
+        memcpy(newMemory, memory(), size());
+        auto newHandle = adoptRef(*new MemoryHandle(newMemory, desiredSize, desiredSize, initial(), maximum(), sharingMode(), MemoryMode::BoundsChecking));
+        m_handle = WTFMove(newHandle);
+
         ASSERT(memory() == newMemory);
         return success();
     }
     case MemoryMode::Signaling: {
-        RELEASE_ASSERT(memory());
+        size_t extraBytes = desiredSize - size();
+        RELEASE_ASSERT(extraBytes);
+        bool allocationSuccess = tryAllocate(
+            [&] () -> MemoryResult::Kind {
+                return memoryManager().tryAllocatePhysicalBytes(extraBytes);
+            }, m_notifyMemoryPressure, m_syncTryToReclaimMemory);
+        if (!allocationSuccess)
+            return makeUnexpected(GrowFailReason::OutOfMemory);
+
+        void* memory = this->memory();
+        RELEASE_ASSERT(memory);
+
         // Signaling memory must have been pre-allocated virtually.
-        uint8_t* startAddress = static_cast<uint8_t*>(memory()) + m_size;
+        uint8_t* startAddress = static_cast<uint8_t*>(memory) + size();
         
-        dataLogLnIf(verbose, "Marking WebAssembly memory's ", RawPointer(memory()), " as read+write in range [", RawPointer(startAddress), ", ", RawPointer(startAddress + extraBytes), ")");
+        dataLogLnIf(verbose, "Marking WebAssembly memory's ", RawPointer(memory), " as read+write in range [", RawPointer(startAddress), ", ", RawPointer(startAddress + extraBytes), ")");
         if (mprotect(startAddress, extraBytes, PROT_READ | PROT_WRITE)) {
             dataLog("mprotect failed: ", strerror(errno), "\n");
             RELEASE_ASSERT_NOT_REACHED();
         }
-        m_memory.recage(m_size, desiredSize);
-        m_size = desiredSize;
+
+        m_handle->growToSize(desiredSize);
         return success();
     }
     }
@@ -447,6 +618,52 @@ Expected<PageCount, Memory::GrowFailReason> Memory::grow(PageCount delta)
     return oldPageCount;
 }
 
+bool Memory::fill(uint32_t offset, uint8_t targetValue, uint32_t count)
+{
+    if (sumOverflows<uint32_t>(offset, count))
+        return false;
+
+    if (offset + count > m_handle->size())
+        return false;
+
+    memset(reinterpret_cast<uint8_t*>(memory()) + offset, targetValue, count);
+    return true;
+}
+
+bool Memory::copy(uint32_t dstAddress, uint32_t srcAddress, uint32_t count)
+{
+    if (sumOverflows<uint32_t>(dstAddress, count) || sumOverflows<uint32_t>(srcAddress, count))
+        return false;
+
+    const uint32_t lastDstAddress = dstAddress + count;
+    const uint32_t lastSrcAddress = srcAddress + count;
+
+    if (lastDstAddress > size() || lastSrcAddress > size())
+        return false;
+
+    if (!count)
+        return true;
+
+    uint8_t* base = reinterpret_cast<uint8_t*>(memory());
+    memcpy(base + dstAddress, base + srcAddress, count);
+    return true;
+}
+
+bool Memory::init(uint32_t offset, const uint8_t* data, uint32_t length)
+{
+    if (sumOverflows<uint32_t>(offset, length))
+        return false;
+
+    if (offset + length > m_handle->size())
+        return false;
+
+    if (!length)
+        return true;
+
+    memcpy(reinterpret_cast<uint8_t*>(memory()) + offset, data, length);
+    return true;
+}
+
 void Memory::registerInstance(Instance* instance)
 {
     size_t count = m_instances.size();
@@ -461,7 +678,8 @@ void Memory::registerInstance(Instance* instance)
 
 void Memory::dump(PrintStream& out) const
 {
-    out.print("Memory at ", RawPointer(memory()), ", size ", m_size, "B capacity ", m_mappedCapacity, "B, initial ", m_initial, " maximum ", m_maximum, " mode ", makeString(m_mode));
+    auto handle = m_handle.copyRef();
+    out.print("Memory at ", RawPointer(handle->memory()), ", size ", handle->size(), "B capacity ", handle->mappedCapacity(), "B, initial ", handle->initial(), " maximum ", handle->maximum(), " mode ", makeString(handle->mode()), " sharingMode ", makeString(handle->sharingMode()));
 }
 
 } // namespace JSC
