@@ -59,23 +59,25 @@ ScrollingStateFrameScrollingNode::ScrollingStateFrameScrollingNode(const Scrolli
     , m_fixedElementsLayoutRelativeToFrame(stateNode.fixedElementsLayoutRelativeToFrame())
     , m_visualViewportIsSmallerThanLayoutViewport(stateNode.visualViewportIsSmallerThanLayoutViewport())
     , m_asyncFrameOrOverflowScrollingEnabled(stateNode.asyncFrameOrOverflowScrollingEnabled())
+    , m_wheelEventGesturesBecomeNonBlocking(stateNode.wheelEventGesturesBecomeNonBlocking())
+    , m_scrollingPerformanceTestingEnabled(stateNode.scrollingPerformanceTestingEnabled())
 {
-    if (hasChangedProperty(RootContentsLayer))
+    if (hasChangedProperty(Property::RootContentsLayer))
         setRootContentsLayer(stateNode.rootContentsLayer().toRepresentation(adoptiveTree.preferredLayerRepresentation()));
 
-    if (hasChangedProperty(CounterScrollingLayer))
+    if (hasChangedProperty(Property::CounterScrollingLayer))
         setCounterScrollingLayer(stateNode.counterScrollingLayer().toRepresentation(adoptiveTree.preferredLayerRepresentation()));
 
-    if (hasChangedProperty(InsetClipLayer))
+    if (hasChangedProperty(Property::InsetClipLayer))
         setInsetClipLayer(stateNode.insetClipLayer().toRepresentation(adoptiveTree.preferredLayerRepresentation()));
 
-    if (hasChangedProperty(ContentShadowLayer))
+    if (hasChangedProperty(Property::ContentShadowLayer))
         setContentShadowLayer(stateNode.contentShadowLayer().toRepresentation(adoptiveTree.preferredLayerRepresentation()));
 
-    if (hasChangedProperty(HeaderLayer))
+    if (hasChangedProperty(Property::HeaderLayer))
         setHeaderLayer(stateNode.headerLayer().toRepresentation(adoptiveTree.preferredLayerRepresentation()));
 
-    if (hasChangedProperty(FooterLayer))
+    if (hasChangedProperty(Property::FooterLayer))
         setFooterLayer(stateNode.footerLayer().toRepresentation(adoptiveTree.preferredLayerRepresentation()));
 }
 
@@ -86,30 +88,35 @@ Ref<ScrollingStateNode> ScrollingStateFrameScrollingNode::clone(ScrollingStateTr
     return adoptRef(*new ScrollingStateFrameScrollingNode(*this, adoptiveTree));
 }
 
-void ScrollingStateFrameScrollingNode::setPropertyChangedBitsAfterReattach()
+OptionSet<ScrollingStateNode::Property> ScrollingStateFrameScrollingNode::applicableProperties() const
 {
-    setPropertyChangedBit(FrameScaleFactor);
-    setPropertyChangedBit(EventTrackingRegion);
-    setPropertyChangedBit(RootContentsLayer);
-    setPropertyChangedBit(ScrolledContentsLayer);
-    setPropertyChangedBit(CounterScrollingLayer);
-    setPropertyChangedBit(InsetClipLayer);
-    setPropertyChangedBit(ContentShadowLayer);
-    setPropertyChangedBit(HeaderHeight);
-    setPropertyChangedBit(FooterHeight);
-    setPropertyChangedBit(HeaderLayer);
-    setPropertyChangedBit(FooterLayer);
-    setPropertyChangedBit(BehaviorForFixedElements);
-    setPropertyChangedBit(TopContentInset);
-    setPropertyChangedBit(FixedElementsLayoutRelativeToFrame);
-    setPropertyChangedBit(VisualViewportIsSmallerThanLayoutViewport);
-    setPropertyChangedBit(AsyncFrameOrOverflowScrollingEnabled);
-    setPropertyChangedBit(LayoutViewport);
-    setPropertyChangedBit(MinLayoutViewportOrigin);
-    setPropertyChangedBit(MaxLayoutViewportOrigin);
-    setPropertyChangedBit(OverrideVisualViewportSize);
+    constexpr OptionSet<Property> nodeProperties = {
+        Property::FrameScaleFactor,
+        Property::EventTrackingRegion,
+        Property::RootContentsLayer,
+        Property::CounterScrollingLayer,
+        Property::InsetClipLayer,
+        Property::ContentShadowLayer,
+        Property::HeaderHeight,
+        Property::FooterHeight,
+        Property::HeaderLayer,
+        Property::FooterLayer,
+        Property::BehaviorForFixedElements,
+        Property::TopContentInset,
+        Property::FixedElementsLayoutRelativeToFrame,
+        Property::VisualViewportIsSmallerThanLayoutViewport,
+        Property::AsyncFrameOrOverflowScrollingEnabled,
+        Property::WheelEventGesturesBecomeNonBlocking,
+        Property::ScrollingPerformanceTestingEnabled,
+        Property::LayoutViewport,
+        Property::MinLayoutViewportOrigin,
+        Property::MaxLayoutViewportOrigin,
+        Property::OverrideVisualViewportSize,
+    };
 
-    ScrollingStateScrollingNode::setPropertyChangedBitsAfterReattach();
+    auto properties = ScrollingStateScrollingNode::applicableProperties();
+    properties.add(nodeProperties);
+    return properties;
 }
 
 void ScrollingStateFrameScrollingNode::setFrameScaleFactor(float scaleFactor)
@@ -119,7 +126,7 @@ void ScrollingStateFrameScrollingNode::setFrameScaleFactor(float scaleFactor)
 
     m_frameScaleFactor = scaleFactor;
 
-    setPropertyChanged(FrameScaleFactor);
+    setPropertyChanged(Property::FrameScaleFactor);
 }
 
 void ScrollingStateFrameScrollingNode::setEventTrackingRegions(const EventTrackingRegions& eventTrackingRegions)
@@ -128,7 +135,7 @@ void ScrollingStateFrameScrollingNode::setEventTrackingRegions(const EventTracki
         return;
 
     m_eventTrackingRegions = eventTrackingRegions;
-    setPropertyChanged(EventTrackingRegion);
+    setPropertyChanged(Property::EventTrackingRegion);
 }
 
 void ScrollingStateFrameScrollingNode::setScrollBehaviorForFixedElements(ScrollBehaviorForFixedElements behaviorForFixed)
@@ -137,7 +144,7 @@ void ScrollingStateFrameScrollingNode::setScrollBehaviorForFixedElements(ScrollB
         return;
 
     m_behaviorForFixed = behaviorForFixed;
-    setPropertyChanged(BehaviorForFixedElements);
+    setPropertyChanged(Property::BehaviorForFixedElements);
 }
 
 void ScrollingStateFrameScrollingNode::setLayoutViewport(const FloatRect& r)
@@ -146,7 +153,7 @@ void ScrollingStateFrameScrollingNode::setLayoutViewport(const FloatRect& r)
         return;
 
     m_layoutViewport = r;
-    setPropertyChanged(LayoutViewport);
+    setPropertyChanged(Property::LayoutViewport);
 }
 
 void ScrollingStateFrameScrollingNode::setMinLayoutViewportOrigin(const FloatPoint& p)
@@ -155,7 +162,7 @@ void ScrollingStateFrameScrollingNode::setMinLayoutViewportOrigin(const FloatPoi
         return;
 
     m_minLayoutViewportOrigin = p;
-    setPropertyChanged(MinLayoutViewportOrigin);
+    setPropertyChanged(Property::MinLayoutViewportOrigin);
 }
 
 void ScrollingStateFrameScrollingNode::setMaxLayoutViewportOrigin(const FloatPoint& p)
@@ -164,16 +171,16 @@ void ScrollingStateFrameScrollingNode::setMaxLayoutViewportOrigin(const FloatPoi
         return;
 
     m_maxLayoutViewportOrigin = p;
-    setPropertyChanged(MaxLayoutViewportOrigin);
+    setPropertyChanged(Property::MaxLayoutViewportOrigin);
 }
 
-void ScrollingStateFrameScrollingNode::setOverrideVisualViewportSize(Optional<FloatSize> viewportSize)
+void ScrollingStateFrameScrollingNode::setOverrideVisualViewportSize(std::optional<FloatSize> viewportSize)
 {
     if (viewportSize == m_overrideVisualViewportSize)
         return;
 
     m_overrideVisualViewportSize = viewportSize;
-    setPropertyChanged(OverrideVisualViewportSize);
+    setPropertyChanged(Property::OverrideVisualViewportSize);
 }
 
 void ScrollingStateFrameScrollingNode::setHeaderHeight(int headerHeight)
@@ -182,7 +189,7 @@ void ScrollingStateFrameScrollingNode::setHeaderHeight(int headerHeight)
         return;
 
     m_headerHeight = headerHeight;
-    setPropertyChanged(HeaderHeight);
+    setPropertyChanged(Property::HeaderHeight);
 }
 
 void ScrollingStateFrameScrollingNode::setFooterHeight(int footerHeight)
@@ -191,7 +198,7 @@ void ScrollingStateFrameScrollingNode::setFooterHeight(int footerHeight)
         return;
 
     m_footerHeight = footerHeight;
-    setPropertyChanged(FooterHeight);
+    setPropertyChanged(Property::FooterHeight);
 }
 
 void ScrollingStateFrameScrollingNode::setTopContentInset(float topContentInset)
@@ -200,7 +207,7 @@ void ScrollingStateFrameScrollingNode::setTopContentInset(float topContentInset)
         return;
 
     m_topContentInset = topContentInset;
-    setPropertyChanged(TopContentInset);
+    setPropertyChanged(Property::TopContentInset);
 }
 
 void ScrollingStateFrameScrollingNode::setRootContentsLayer(const LayerRepresentation& layerRepresentation)
@@ -209,7 +216,7 @@ void ScrollingStateFrameScrollingNode::setRootContentsLayer(const LayerRepresent
         return;
 
     m_rootContentsLayer = layerRepresentation;
-    setPropertyChanged(RootContentsLayer);
+    setPropertyChanged(Property::RootContentsLayer);
 }
 
 void ScrollingStateFrameScrollingNode::setCounterScrollingLayer(const LayerRepresentation& layerRepresentation)
@@ -218,7 +225,7 @@ void ScrollingStateFrameScrollingNode::setCounterScrollingLayer(const LayerRepre
         return;
     
     m_counterScrollingLayer = layerRepresentation;
-    setPropertyChanged(CounterScrollingLayer);
+    setPropertyChanged(Property::CounterScrollingLayer);
 }
 
 void ScrollingStateFrameScrollingNode::setInsetClipLayer(const LayerRepresentation& layerRepresentation)
@@ -227,7 +234,7 @@ void ScrollingStateFrameScrollingNode::setInsetClipLayer(const LayerRepresentati
         return;
     
     m_insetClipLayer = layerRepresentation;
-    setPropertyChanged(InsetClipLayer);
+    setPropertyChanged(Property::InsetClipLayer);
 }
 
 void ScrollingStateFrameScrollingNode::setContentShadowLayer(const LayerRepresentation& layerRepresentation)
@@ -236,7 +243,7 @@ void ScrollingStateFrameScrollingNode::setContentShadowLayer(const LayerRepresen
         return;
     
     m_contentShadowLayer = layerRepresentation;
-    setPropertyChanged(ContentShadowLayer);
+    setPropertyChanged(Property::ContentShadowLayer);
 }
 
 void ScrollingStateFrameScrollingNode::setHeaderLayer(const LayerRepresentation& layerRepresentation)
@@ -245,7 +252,7 @@ void ScrollingStateFrameScrollingNode::setHeaderLayer(const LayerRepresentation&
         return;
     
     m_headerLayer = layerRepresentation;
-    setPropertyChanged(HeaderLayer);
+    setPropertyChanged(Property::HeaderLayer);
 }
 
 void ScrollingStateFrameScrollingNode::setFooterLayer(const LayerRepresentation& layerRepresentation)
@@ -254,7 +261,7 @@ void ScrollingStateFrameScrollingNode::setFooterLayer(const LayerRepresentation&
         return;
     
     m_footerLayer = layerRepresentation;
-    setPropertyChanged(FooterLayer);
+    setPropertyChanged(Property::FooterLayer);
 }
 
 void ScrollingStateFrameScrollingNode::setVisualViewportIsSmallerThanLayoutViewport(bool visualViewportIsSmallerThanLayoutViewport)
@@ -263,7 +270,7 @@ void ScrollingStateFrameScrollingNode::setVisualViewportIsSmallerThanLayoutViewp
         return;
     
     m_visualViewportIsSmallerThanLayoutViewport = visualViewportIsSmallerThanLayoutViewport;
-    setPropertyChanged(VisualViewportIsSmallerThanLayoutViewport);
+    setPropertyChanged(Property::VisualViewportIsSmallerThanLayoutViewport);
 }
 
 void ScrollingStateFrameScrollingNode::setFixedElementsLayoutRelativeToFrame(bool fixedElementsLayoutRelativeToFrame)
@@ -272,7 +279,7 @@ void ScrollingStateFrameScrollingNode::setFixedElementsLayoutRelativeToFrame(boo
         return;
     
     m_fixedElementsLayoutRelativeToFrame = fixedElementsLayoutRelativeToFrame;
-    setPropertyChanged(FixedElementsLayoutRelativeToFrame);
+    setPropertyChanged(Property::FixedElementsLayoutRelativeToFrame);
 }
 
 void ScrollingStateFrameScrollingNode::setAsyncFrameOrOverflowScrollingEnabled(bool enabled)
@@ -281,7 +288,25 @@ void ScrollingStateFrameScrollingNode::setAsyncFrameOrOverflowScrollingEnabled(b
         return;
     
     m_asyncFrameOrOverflowScrollingEnabled = enabled;
-    setPropertyChanged(AsyncFrameOrOverflowScrollingEnabled);
+    setPropertyChanged(Property::AsyncFrameOrOverflowScrollingEnabled);
+}
+
+void ScrollingStateFrameScrollingNode::setWheelEventGesturesBecomeNonBlocking(bool enabled)
+{
+    if (enabled == m_wheelEventGesturesBecomeNonBlocking)
+        return;
+    
+    m_wheelEventGesturesBecomeNonBlocking = enabled;
+    setPropertyChanged(Property::WheelEventGesturesBecomeNonBlocking);
+}
+
+void ScrollingStateFrameScrollingNode::setScrollingPerformanceTestingEnabled(bool enabled)
+{
+    if (enabled == m_scrollingPerformanceTestingEnabled)
+        return;
+    
+    m_scrollingPerformanceTestingEnabled = enabled;
+    setPropertyChanged(Property::ScrollingPerformanceTestingEnabled);
 }
 
 void ScrollingStateFrameScrollingNode::dumpProperties(TextStream& ts, ScrollingStateTreeAsTextBehavior behavior) const
@@ -292,11 +317,16 @@ void ScrollingStateFrameScrollingNode::dumpProperties(TextStream& ts, ScrollingS
     
     if (behavior & ScrollingStateTreeAsTextBehaviorIncludeLayerIDs) {
         ts.dumpProperty("root contents layer ID", m_rootContentsLayer.layerID());
-        ts.dumpProperty("counter scrolling layer ID", m_counterScrollingLayer.layerID());
-        ts.dumpProperty("inset clip layer ID", m_insetClipLayer.layerID());
-        ts.dumpProperty("content shadow layer ID", m_contentShadowLayer.layerID());
-        ts.dumpProperty("header layer ID", m_headerLayer.layerID());
-        ts.dumpProperty("footer layer ID", m_footerLayer.layerID());
+        if (m_counterScrollingLayer.layerID())
+            ts.dumpProperty("counter scrolling layer ID", m_counterScrollingLayer.layerID());
+        if (m_insetClipLayer.layerID())
+            ts.dumpProperty("inset clip layer ID", m_insetClipLayer.layerID());
+        if (m_contentShadowLayer.layerID())
+            ts.dumpProperty("content shadow layer ID", m_contentShadowLayer.layerID());
+        if (m_headerLayer.layerID())
+            ts.dumpProperty("header layer ID", m_headerLayer.layerID());
+        if (m_footerLayer.layerID())
+            ts.dumpProperty("footer layer ID", m_footerLayer.layerID());
     }
 
     if (m_frameScaleFactor != 1)
@@ -315,9 +345,6 @@ void ScrollingStateFrameScrollingNode::dumpProperties(TextStream& ts, ScrollingS
     if (m_overrideVisualViewportSize)
         ts.dumpProperty("override visual viewport size", m_overrideVisualViewportSize.value());
 
-    if (m_behaviorForFixed == StickToViewportBounds)
-        ts.dumpProperty("behavior for fixed", m_behaviorForFixed);
-
     if (!m_eventTrackingRegions.asynchronousDispatchRegion.isEmpty()) {
         TextStream::GroupScope scope(ts);
         ts << "asynchronous event dispatch region";
